# Copyright (c) 2017 T-semi All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer. 
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY T-SEMI "AS IS" AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL T-SEMI BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

package KueChip2;

use warnings;
use strict;
use utf8;

sub new{
    my $class = shift;
    my @mim;
    for(0x000 .. 0x1ff){
        my $mv = shift || 0;
        push @mim, $mv & 0xff;
    }
    return bless {
            mem => [@mim],
            acc => 0,
            ix  => 0,
            pc  => 0,
            cf  => 0,
            vf  => 0,
            nf  => 0,
            zf  => 0,
            run => 0,
            stepcnt => 0,
            lastset => undef,
            lastop  => undef
        }, $class;
}

sub dump{
    my $self = shift;
    for(0x00 .. 0x1f){
        printf("%02x0 ", $_);
        for(($_ << 4) .. ($_ << 4) + 0x0f){
            printf("%02x ", $self->{mem}[$_]);
        }
        print("\n");
    }
    printf("acc: %02x, ix: %02x, flag: %02x, pc: %02x\n",
            $self->acc, $self->ix, $self->flag, $self->pc);
}

sub acc{
    my $self = shift;
    my $val = shift;
    if(defined($val)){
        $self->{acc} = $val & 0xff;
    }else{
        return $self->{acc};
    }
}

sub ix{
    my $self = shift;
    my $val = shift;
    if(defined($val)){
        $self->{ix} = $val & 0xff;
    }else{
        return $self->{ix};
    }
}

sub pc{
    my $self = shift;
    my $val = shift;
    if(defined($val)){
        $self->{pc} = $val & 0xff;
    }else{
        return $self->{pc};
    }
}

sub flag{
    my $self = shift;
    if(@_ == 1){
        my $f = shift;
        $self->{cf} = !!($f & 0x08);
        $self->{vf} = !!($f & 0x04);
        $self->{nf} = !!($f & 0x02);
        $self->{zf} = !!($f & 0x01);
    }elsif(@_){
        $self->{cf} = !!shift;
        $self->{vf} = !!shift;
        $self->{nf} = !!shift;
        $self->{zf} = !!shift;
    }else{
        return ($self->{cf}<<3)|($self->{vf}<<2)|
               ($self->{nf}<<1)|($self->{zf}<<0);
    }
}

sub running{
    my $self = shift;
    return $self->{run};
}

sub halt{
    my $self = shift;
    $self->{run} = 0;
}

sub run{
    my $self = shift;
    $self->{run} = 1;
}

sub mmap{
    my $self = shift;
    if(@_){
        $self->{mem} = [@_];
    }else{
        return @{$self->{mem}};
    }
}

sub stepcnt{
    my $self = shift;
    my $val = shift;
    if(defined($val)){
        $self->{stepcnt} = $val;
    }else{
        return $self->{stepcnt};
    }
}

sub with_sign{
    my $self = shift;
    my $val = shift;
    if($val & 0x80){
        $val = -((~$val + 1)&0xff);
    }
    return $val;
}

sub get_value{
    my $self = shift;
    my $opr = lc(shift);
    if($opr eq "acc"){
        return $self->{acc};
    }elsif($opr eq "ix"){
        return $self->{ix};
    }elsif($opr =~ /^\+?([0-9a-f]+)h$/ || $opr =~ /^\+?0x([0-9a-f]+)$/){
        return hex($1);
    }elsif($opr =~ /^\+?([0-9]+)$/){
        return $1;
    }elsif($opr =~ /^-([0-9a-f]+)h$/ || $opr =~ /^-0x([0-9a-f]+)$/){
        return -hex($1);
    }elsif($opr =~ /^-([0-9]+)$/){
        return -$1;
    }elsif($opr =~ /^\[(.*)(\])$/ || $opr =~ /^\((.*)(\))$/){
        my $idx = $2 eq "]" ? 0 : 0x100;
        my $i = $1;
        if($i =~ /^([0-9a-f]+)h$/ || $i =~ /^0x([0-9a-f]+)$/){
            return $self->{mem}[(hex($1)&0xff)+$idx];
        }elsif($i =~ /^([0-9]+)$/){
            return $self->{mem}[($1&0xff)+$idx];
        }elsif($i =~ /^ix\+([0-9a-f]+)h$/ || $i =~ /^ix\+0x([0-9a-f]+)$/){
            return $self->{mem}[($self->{ix}+hex($1) & 0xff)+$idx];
        }elsif($i =~ /^ix\+([0-9]+)$/){
            return $self->{mem}[($self->{ix}+$1 & 0xff)+$idx];
        }elsif($i =~ /^ix-([0-9a-f]+)h$/ || $i =~ /^ix-0x([0-9a-f]+)$/){
            return $self->{mem}[($self->{ix}-hex($1) & 0xff)+$idx];
        }elsif($i =~ /^ix-([0-9]+)$/){
            return $self->{mem}[($self->{ix}-$1 & 0xff)+$idx];
        }
    }
    die;
}

sub get_lastchange{
    my $self = shift;
    if(!defined($self->{lastset})){
        return ();
    }elsif($self->{lastset} =~ /^\d+$/){
        return ($self->{lastset}, $self->{mem}[$self->{lastset}]);
    }else{
        return ($self->{lastset}, $self->{$self->{lastset}});
    }
}

sub get_lastop{
    my $self = shift;
    return @{$self->{lastop}};
}

sub set_value{
    my $self = shift;
    my $opr = lc(shift);
    my $val = shift;
    $val &= 0xff;
    if($opr eq "acc"){
        $self->{lastset} = "acc";
        return $self->{acc} = $val;
    }elsif($opr eq "ix"){
        $self->{lastset} = "ix";
        return $self->{ix} = $val;
    }elsif($opr =~ /^\[(.*)(\])$/ || $opr =~ /^\((.*)(\))$/){
        my $idx = $2 eq "]" ? 0 : 0x100;
        my $i = $1;
        if($i =~ /^([0-9a-f]+)h$/ || $i =~ /^0x([0-9a-f]+)$/){
            $self->{lastset} = (hex($1)&0xff)+$idx;
            return $self->{mem}[$self->{lastset}] = $val;
        }elsif($i =~ /^([0-9]+)$/){
            $self->{lastset} = ($1&0xff)+$idx;
            return $self->{mem}[$self->{lastset}] = $val;
        }elsif($i =~ /^ix\+([0-9a-f]+)h$/ || $i =~ /^ix\+0x([0-9a-f]+)$/){
            $self->{lastset} = ($self->{ix}+hex($1) & 0xff)+$idx;
            return $self->{mem}[$self->{lastset}] = $val;
        }elsif($i =~ /^ix\+([0-9]+)$/){
            $self->{lastset} = ($self->{ix}+$1 & 0xff)+$idx;
            return $self->{mem}[$self->{lastset}] = $val;
        }elsif($i =~ /^ix-([0-9a-f]+)h$/ || $i =~ /^ix-0x([0-9a-f]+)$/){
            $self->{lastset} = ($self->{ix}-hex($1) & 0xff)+$idx;
            return $self->{mem}[$self->{lastset}] = $val;
        }elsif($i =~ /^ix-([0-9]+)$/){
            $self->{lastset} = ($self->{ix}-$1 & 0xff)+$idx;
            return $self->{mem}[$self->{lastset}] = $val;
        }
    }
    die;
}

sub readpc{
    my $self = shift;
    my $rt = $self->{mem}[$self->{pc}];
    $self->{pc}++;
    $self->{pc} &= 0xff;
    return $rt;
}

sub dostep{
    my $self = shift;
    $self->{lastset} = undef;
    $self->{run} = 1;
    my @op = $self->fetchop();
    $self->{lastop} = \@op;
    $self->do_instruction(@op);
    return $self->{run};
}

sub dowithlog{
    my $self = shift;
    my @bps = @_;
    my $logger = sub{
        my $lm = "";
        my @lc = $self->get_lastchange();
        if(@lc){
            if($lc[0] =~ /^\d+$/){
                $lm .= sprintf("%03xh=%02xh, ", $lc[0], $lc[1]);
            }else{
                $lm .= sprintf("%s=%02xh, ", $lc[0], $lc[1]);
            }
        }
        $lm .= sprintf("pc:%02xh, flg:", $self->{pc});
        for(qw/c v n z/){
            if($self->{$_."f"}){
                $lm .= $_;
            }else{
                $lm .= " ";
            }
        }
        print "$lm\n";
    };
    while(1){
        my @op = $self->fetchop();
        print "operating:".join(",", @op)."\n";
        $self->{lastop} = \@op;
        $self->{lastset} = undef;
        $self->{run} = 1;
        $self->do_instruction(@op);
        if(!$self->{run}){
            last;
        }
        for(@bps){
            $self->dump() if($_ == $self->{pc});
        }
        $logger->();
    }
    $logger->();
}

sub do{
    my $self = shift;
    my @bps = @_;
    while($self->dostep()){
        for(@bps){
            return if($_ == $self->{pc});
        }
    }
}

sub gimm{
    my $self = shift;
    my $v = shift;
    if($v =~ /^\s*([+-]?)\s*0x([0-9a-f]+)\s*$/i ||
       $v =~ /^\s*([+-]?)\s*([0-9a-f]+)h\s*$/i){
        return $1 eq "-" ? -hex($2) : hex($2);
    }elsif($v =~ /^\s*([+-]?)\s*([0-9]+)\s*$/){
        return $1 eq "-" ? -$2 : $2;
    }elsif($v =~ /^\s*([+-]?[^+-]+)
                      ([+-] .+)$/x){
        return ($self->gimm($1), $self->gimm($2));
    }elsif($v =~ /^\s*\+?([^\s]+)\s*$/){
        return $1;
    }else{
        die;
    }
}

sub asm{
    my $self = shift;
    my @asmcode;
    @_ || die;
    if(@_ == 1){
        @asmcode = split(/\n/, shift);
    }else{
        @asmcode = @_;
    }

    my $caddr = 0;
    my %labels;
    my @lplaces;
    for(@asmcode){
        if(/^\s*
            (?: ([A-Za-z_][A-Za-z_0-9]*) \s* : \s* )?
            (?:
                ([A-Za-z]+) \s*
                (?: \s+
                    ([^#,]+) \s*
                    (?: , \s* ([^#,]+) )? \s*
                )?
            )?
            (?: \#.* )?
        $/x){
            my $label = $1;
            my $opname = $2;
            my $opr1 = $3;
            my $opr2 = $4;
            if(defined($label)){
                if(defined($opname) && lc($opname) eq "equ"){
                    die if(!defined($opr1) || defined($opr2));
                    my @addr = ($self->gimm($opr1));
                    if(@addr > 1 || $addr[0] !~ /^[0-9]+$/){
                        die;
                    }
                    $labels{$label} = $addr[0];
                }else{
                    $labels{$label} = $caddr;
                }
            }
            if(defined($opname)){
                my @ld_fmly = qw/ld st sbc adc sub add eor or and cmp/;
                my @shift_fmly = qw/sra sla srl sll rra rla rrl rll/;
                my @branch_fmly = (qw/ba bnz bzp bp bni bnc bge bgt/,
                                   qw/bvf bz bn bzn bno bc blt ble/);
                my %others = (
                    "nop" => 0x00,
                    "hlt" => 0x0f,
                    "out" => 0x10,
                    "in"  => 0x1f,
                    "rcf" => 0x20,
                    "scf" => 0x2f
                );
                $opname = lc($opname);
                if(grep{/^$opname$/}(@ld_fmly)){
                    my ($opn) = grep{$ld_fmly[$_]eq$opname}(0 .. $#ld_fmly);
                    $opn = ($opn + 6) << 4;
                    die if(!defined($opr1) || !defined($opr2));
                    $opr1 =~ s/\s//g;
                    $opr1 = lc($opr1);
                    if($opr1 eq "ix"){
                        $opn |= 0x08;
                    }elsif($opr1 ne "acc"){
                        die;
                    }

                    my $bopr = undef;
                    if($opr2 =~ /^\s*acc\s*$/i){
                        $opn |= 0;
                    }elsif($opr2 =~ /^\s*ix\s*$/i){
                        $opn |= 1;
                    }elsif($opr2 =~ /^\s*\[ix\s*([+-].+)\]\s*$/i){
                        $opn |= 6;
                        $bopr = [$self->gimm($1)];
                    }elsif($opr2 =~ /^\s*\(ix\s*([+-].+)\)\s*$/i){
                        $opn |= 7;
                        $bopr = [$self->gimm($1)];
                    }elsif($opr2 =~ /^\s*\[(.+)\]\s*$/){
                        $opn |= 4;
                        $bopr = [$self->gimm($1)];
                    }elsif($opr2 =~ /^\s*\((.+)\)\s*$/){
                        $opn |= 5;
                        $bopr = [$self->gimm($1)];
                    }else{
                        $opn |= 2;
                        $bopr = [$self->gimm($opr2)];
                    }

                    $self->{mem}[$caddr++] = $opn;
                    if($bopr){
                        push @lplaces, {addr=>$caddr++, val=>$bopr};
                    }
                }elsif(grep{/^$opname$/}(@shift_fmly)){
                    die if(!defined($opr1) || defined($opr2));
                    my ($opn) = grep{$shift_fmly[$_]eq$opname}
                                        (0 .. $#shift_fmly);
                    $opn |= 0x40;
                    $opr1 =~ s/\s//g;
                    $opr1 = lc($opr1);
                    if($opr1 eq "ix"){
                        $opn |= 0x08;
                    }elsif($opr1 ne "acc"){
                        die;
                    }
                    $self->{mem}[$caddr++] = $opn;
                }elsif(grep{/^$opname$/}(@branch_fmly)){
                    die if(!defined($opr1) || defined($opr2));
                    my ($opn) = grep{$branch_fmly[$_]eq$opname}
                                        (0 .. $#branch_fmly);
                    $opn |= 0x30;
                    $self->{mem}[$caddr++] = $opn;
                    push @lplaces, {addr=>$caddr++,
                                    val=>[$self->gimm($opr1)]};
                }elsif(defined($others{$opname})){
                    $self->{mem}[$caddr++] = $others{$opname};
                }elsif($opname ne "equ"){
                    die;
                }
            }
        }else{
            die;
        }
    }

    die if($caddr > 0x100);

    for(@lplaces){
        my $addr = $_->{addr};
        my $val = 0;
        for(@{$_->{val}}){
            if(/^-?\d+$/){
                $val += $_;
            }elsif(/^-(.+)$/){
                $val -= $labels{$1};
            }elsif(defined($labels{$_})){
                $val += $labels{$_};
            }else{
                die;
            }
        }
        $self->{mem}[$addr] = $val & 0xff;
    }
}

sub fetchop{
    my $self = shift;
    my $numfmt = shift || "%d";
    my $opc = $self->readpc();
    if(($opc & 0xf0) >= 0x60){
        my $op = ${[qw/ld st sbc adc sub add eor or and cmp/]}[($opc>>4)-6];
        my $dst = $opc & 0x08 ? "ix" : "acc";
        my $src = $opc & 0x07;
        if($src == 0){
            $src = "acc";
        }elsif($src == 1){
            $src = "ix";
        }elsif($src == 2){
            $src = sprintf($numfmt, $self->readpc());
        }elsif($src == 4){
            $src = sprintf("[$numfmt]", $self->readpc());
        }elsif($src == 5){
            $src = sprintf("($numfmt)", $self->readpc());
        }elsif($src == 6){
            $src = sprintf("[ix+$numfmt]", $self->readpc());
        }elsif($src == 7){
            $src = sprintf("(ix+$numfmt)", $self->readpc());
        }else{
            die;
        }
        return ($op, $dst, $src);
    }elsif(($opc & 0xf0) == 0x40){
        my $op = ${[qw/sra sla srl sll rra rla rrl rll/]}[$opc&0x07];
        my $dst = $opc & 0x08 ? "ix" : "acc";
        return ($op, $dst);
    }elsif(($opc & 0xf0) == 0x30){
        my $op = ${[qw/ba bnz bzp bp bni bnc bge bgt/,
                    qw/bvf bz bn bzn bno bc blt ble/]}[$opc&0x0f];
        my $to = sprintf($numfmt, $self->readpc());
        return ($op, $to);
    }elsif($opc == 0x00){
        return "nop";
    }elsif($opc == 0x0f){
        return "hlt";
    }elsif($opc == 0x10){
        return "out";
    }elsif($opc == 0x1f){
        return "in";
    }elsif($opc == 0x20){
        return "rcf";
    }elsif($opc == 0x2f){
        return "scf";
    }else{
        die sprintf("unknown op %02x", $opc);
    }
}

sub do_instruction{
    my $self = shift;
    my $opc = lc(shift);
    my $op1 = shift;
    my $op2 = shift;
    $self->{stepcnt}++;
    if($opc eq "ld"){
        if($op1 eq "acc" || $op1 eq "ix"){
            $self->set_value($op1, $self->get_value($op2));
        }else{
            die;
        }
        return;
    }elsif($opc eq "st"){
        if($op2 eq "acc" || $op2 eq "ix"){
            die;
        }
        if($op1 eq "acc" || $op1 eq "ix"){
            $self->set_value($op2, $self->get_value($op1));
        }else{
            die;
        }
        return;
    }elsif($opc eq "sbc"){
        my $v1 = $self->get_value($op1);
        my $v2 = $self->get_value($op2);
        my $ans = $v1 - $v2 - $self->{cf};
        my $aws = $self->with_sign($v1) - $self->with_sign($v2)
                                        - $self->{cf};
        $self->{vf} = $aws < -128 || $aws > 127;
        $self->{cf} = $ans < 0;
        $ans &= 0xff;
        $self->{nf} = !!($ans & 0x80);
        $self->{zf} = !$ans;
        $self->set_value($op1, $ans);
        return;
    }elsif($opc eq "adc"){
        my $v1 = $self->get_value($op1);
        my $v2 = $self->get_value($op2);
        my $ans = $v1 + $v2 + $self->{cf};
        my $aws = $self->with_sign($v1) + $self->with_sign($v2)
                                        + $self->{cf};
        $self->{vf} = $aws < -128 || $aws > 127;
        $self->{cf} = $ans > 0xff;
        $ans &= 0xff;
        $self->{nf} = !!($ans & 0x80);
        $self->{zf} = !$ans;
        $self->set_value($op1, $ans);
        return;
    }elsif($opc eq "sub"){
        my $v1 = $self->get_value($op1);
        my $v2 = $self->get_value($op2);
        my $ans = $self->with_sign($v1) - $self->with_sign($v2);
        $self->{vf} = $ans < -128 || $ans > 127;
        $ans &= 0xff;
        $self->{nf} = !!($ans & 0x80);
        $self->{zf} = !$ans;
        $self->set_value($op1, $ans);
        return;
    }elsif($opc eq "add"){
        my $v1 = $self->get_value($op1);
        my $v2 = $self->get_value($op2);
        my $ans = $self->with_sign($v1) + $self->with_sign($v2);
        $self->{vf} = $ans < -128 || $ans > 127;
        $ans &= 0xff;
        $self->{nf} = !!($ans & 0x80);
        $self->{zf} = !$ans;
        $self->set_value($op1, $ans);
        return;
    }elsif($opc eq "eor"){
        my $v1 = $self->get_value($op1);
        my $v2 = $self->get_value($op2);
        my $ans = $v1 ^ $v2;
        $self->{vf} = 0;
        $self->{nf} = !!($ans & 0x80);
        $self->{zf} = !$ans;
        $self->set_value($op1, $ans);
        return;
    }elsif($opc eq "or"){
        my $v1 = $self->get_value($op1);
        my $v2 = $self->get_value($op2);
        my $ans = $v1 | $v2;
        $self->{vf} = 0;
        $self->{nf} = !!($ans & 0x80);
        $self->{zf} = !$ans;
        $self->set_value($op1, $ans);
        return;
    }elsif($opc eq "and"){
        my $v1 = $self->get_value($op1);
        my $v2 = $self->get_value($op2);
        my $ans = $v1 & $v2;
        $self->{vf} = 0;
        $self->{nf} = !!($ans & 0x80);
        $self->{zf} = !$ans;
        $self->set_value($op1, $ans);
        return;
    }elsif($opc eq "cmp"){
        my $v1 = $self->get_value($op1);
        my $v2 = $self->get_value($op2);
        my $ans = $self->with_sign($v1) - $self->with_sign($v2);
        $self->{vf} = $ans < -128 || $ans > 127;
        $ans &= 0xff;
        $self->{nf} = !!($ans & 0x80);
        $self->{zf} = !$ans;
        return;
    }
    die if(defined($op2));
    if($opc eq "sra"){
        my $ans = $self->get_value($op1);
        $self->{cf} = !!($ans & 0x01);
        $ans = ($ans >> 1) & 0x7f | ($ans & 0x80);
        $self->{vf} = 0;
        $self->{nf} = !!($ans & 0x80);
        $self->{zf} = !$ans;
        $self->set_value($op1, $ans);
        return;
    }elsif($opc eq "sla"){
        my $ans = $self->get_value($op1);
        $self->{cf} = !!($ans & 0x80);
        $ans = ($ans << 1) & 0xff;
        $self->{vf} = $self->{cf};
        $self->{nf} = !!($ans & 0x80);
        $self->{zf} = !$ans;
        $self->set_value($op1, $ans);
        return;
    }elsif($opc eq "srl"){
        my $ans = $self->get_value($op1);
        $self->{cf} = !!($ans & 0x01);
        $ans = ($ans >> 1) & 0xff;
        $self->{vf} = 0;
        $self->{nf} = !!($ans & 0x80);
        $self->{zf} = !$ans;
        $self->set_value($op1, $ans);
        return;
    }elsif($opc eq "sll"){
        my $ans = $self->get_value($op1);
        $self->{cf} = !!($ans & 0x80);
        $ans = ($ans << 1) & 0xff;
        $self->{vf} = 0;
        $self->{nf} = !!($ans & 0x80);
        $self->{zf} = !$ans;
        $self->set_value($op1, $ans);
        return;
    }elsif($opc eq "rra"){
        my $v = $self->get_value($op1);
        my $ans = ($v >> 1) & 0x7f | ($self->{cf} << 7);
        $self->{cf} = !!($v & 0x01);
        $self->{vf} = 0;
        $self->{nf} = !!($ans & 0x80);
        $self->{zf} = !$ans;
        $self->set_value($op1, $ans);
        return;
    }elsif($opc eq "rla"){
        my $v = $self->get_value($op1);
        my $ans = ($v << 1) & 0xff | $self->{cf};
        $self->{cf} = !!($v & 0x80);
        $self->{vf} = $self->{cf};
        $self->{nf} = !!($ans & 0x80);
        $self->{zf} = !$ans;
        $self->set_value($op1, $ans);
        return;
    }elsif($opc eq "rrl"){
        my $ans = $self->get_value($op1);
        $self->{cf} = !!($ans & 0x01);
        $ans = ($ans >> 1) & 0xff | ($self->{cf} << 7);
        $self->{vf} = 0;
        $self->{nf} = !!($ans & 0x80);
        $self->{zf} = !$ans;
        $self->set_value($op1, $ans);
        return;
    }elsif($opc eq "rll"){
        my $ans = $self->get_value($op1);
        $self->{cf} = !!($ans & 0x80);
        $ans = ($ans << 1) & 0xff | $self->{cf};
        $self->{vf} = 0;
        $self->{nf} = !!($ans & 0x80);
        $self->{zf} = !$ans;
        $self->set_value($op1, $ans);
        return;
    }elsif($opc eq "ba"){
        $self->{pc} = $self->get_value($op1);
        return;
    }elsif($opc eq "bvf"){
        $self->{pc} = $self->get_value($op1) if($self->{vf});
        return;
    }elsif($opc eq "bnz"){
        $self->{pc} = $self->get_value($op1) if(!$self->{zf});
        return;
    }elsif($opc eq "bz"){
        $self->{pc} = $self->get_value($op1) if($self->{zf});
        return;
    }elsif($opc eq "bzp"){
        $self->{pc} = $self->get_value($op1) if(!$self->{nf});
        return;
    }elsif($opc eq "bn"){
        $self->{pc} = $self->get_value($op1) if($self->{nf});
        return;
    }elsif($opc eq "bp"){
        $self->{pc} = $self->get_value($op1) if(!($self->{nf}||$self->{zf}));
        return;
    }elsif($opc eq "bzn"){
        $self->{pc} = $self->get_value($op1) if($self->{nf}||$self->{zf});
        return;
    }elsif($opc eq "bni"){
        ...
    }elsif($opc eq "bno"){
        ...
    }elsif($opc eq "bnc"){
        $self->{pc} = $self->get_value($op1) if(!$self->{cf});
        return;
    }elsif($opc eq "bc"){
        $self->{pc} = $self->get_value($op1) if($self->{cf});
        return;
    }elsif($opc eq "bge"){
        $self->{pc} = $self->get_value($op1) if(!($self->{vf}^$self->{nf}));
        return;
    }elsif($opc eq "blt"){
        $self->{pc} = $self->get_value($op1) if($self->{vf}^$self->{nf});
        return;
    }elsif($opc eq "bgt"){
        $self->{pc} = $self->get_value($op1)
                    if(!(($self->{vf}^$self->{nf})||$self->{zf}));
        return;
    }elsif($opc eq "ble"){
        $self->{pc} = $self->get_value($op1)
                    if(($self->{vf}^$self->{nf})||$self->{zf});
        return;
    }
    die if(defined($op1));
    if($opc eq "nop"){
        return;
    }elsif($opc eq "hlt"){
        $self->{run} = 0;
        return;
    }elsif($opc eq "out"){
        ...
    }elsif($opc eq "in"){
        ...
    }elsif($opc eq "rcf"){
        $self->{cf} = 0;
        return;
    }elsif($opc eq "scf"){
        $self->{cf} = 1;
        return;
    }
    die;
}

1;
