#!/usr/bin/perl

# Copyright (c) 2017 T-semi All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer. 
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY T-SEMI "AS IS" AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL T-SEMI BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

use warnings;
use strict;
use utf8;
use Data::Dumper;

my $st_sign = "←";
my $shorten_shift = 1;
my $hide_compare = 1;
my $box_width = 70;
my $box_height = 24;
my $box_dist = 6;
my $access_dist = 10;
my $access_width = 50;
my $jumpline_offset = 10;
my $jump_preline = 2;
my $jumpline_dist = 20;
my $text_size = 8;
my $text_font = "VL ゴシック";

binmode STDOUT, ":utf8";
binmode STDIN, ":utf8";

sub toxmlstr{
    my $t = shift;
    $t =~ s/&/&amp;/g;
    $t =~ s/</&lt;/g;
    $t =~ s/>/&gt;/g;
    $t =~ s/'/&apos;/g;
    $t =~ s/"/&quot;/g;
    return $t;
}

sub svg_tex{
    my $text = shift;
    my $x = shift;
    my $y = shift;
    my $maxwidth = shift;

    my $width = 0;
    for(split(//, $text)){
        if(ord($_) <= 0x7f){
            $width += 1;
        }else{
            $width += 2;
        }
    }
    $width = $width / 2 * $text_size;
    if($width > $maxwidth){
        $width = $maxwidth;
    }
    
    return "<text x=\"$x\" y=\"$y\" "
         . "font-family=\"$text_font\" font-size=\"$text_size\" "
         . "fill=\"black\" stroke=\"none\" textLength=\"$width\" "
         . "lengthAdjust=\"spacingAndGlyphs\" "
         . "text-anchor=\"middle\" dominant-baseline=\"middle\">"
         . toxmlstr($text)
         . "</text>";
}

sub gimm{
    my $v = shift;
    if($v =~ /^\s*([+-]?)\s*0x([0-9a-f]+)\s*$/i ||
       $v =~ /^\s*([+-]?)\s*([0-9a-f]+)h\s*$/i){
        return $1 eq "-" ? -hex($2) : hex($2);
    }elsif($v =~ /^\s*([+-]?)\s*([0-9]+)\s*$/){
        return $1 eq "-" ? -$2 : $2;
    }elsif($v =~ /^\s*([+-]?[^+-]+)
                      ([+-] .+)$/x){
        return (gimm($1), gimm($2));
    }elsif($v =~ /^\s*\+?([^\s]+)\s*$/){
        return $1;
    }else{
        die;
    }
}

sub asm{
    my @asmcode;
    @_ || die;
    if(@_ == 1){
        @asmcode = split(/\n/, shift);
    }else{
        @asmcode = @_;
    }

    my $caddr = 0;
    my %labels;
    my @opr_list;
    for(@asmcode){
        if(/^\s*
            (?: ([A-Za-z_][A-Za-z_0-9]*) \s* : \s* )?
            (?:
                ([A-Za-z]+) \s*
                (?: \s+
                    ([^#,\s]+) \s*
                    (?: , \s* ([^#,\s]+) )? \s*
                )?
            )?
            (?: \#.* )?
        $/x){
            my $label = $1;
            my $opname = $2;
            my $opr1 = $3;
            my $opr2 = $4;
            if(defined($label)){
                if(defined($opname) && lc($opname) eq "equ"){
                    die if(!defined($opr1) || defined($opr2));
                    my @addr = (gimm($opr1));
                    if(@addr > 1 || $addr[0] !~ /^[0-9]+$/){
                        die;
                    }
                    $addr[0] &= 0xff;
                    $labels{$label} = {on=>undef, addr=>$addr[0]};
                }else{
                    $labels{$label} = {on=>int(@opr_list), addr=>$caddr};
                }
            }
            if(defined($opname)){
                my @ld_fmly = qw/ld st sbc adc sub add eor or and cmp/;
                my @shift_fmly = qw/sra sla srl sll rra rla rrl rll/;
                my @branch_fmly = (qw/ba bnz bzp bp bni bnc bge bgt/,
                                   qw/bvf bz bn bzn bno bc blt ble/);
                my %others = (
                    "nop" => 0x00,
                    "hlt" => 0x0f,
                    "out" => 0x10,
                    "in"  => 0x1f,
                    "rcf" => 0x20,
                    "scf" => 0x2f
                );
                $opname = lc($opname);
                if(grep{/^$opname$/}(@ld_fmly)){
                    die if(!defined($opr1) || !defined($opr2));
                    $opr1 =~ s/\s//g;
                    $opr1 = lc($opr1);
                    if($opr1 ne "acc" && $opr1 ne "ix"){
                        die;
                    }

                    my $ix_flg = 0;
                    my $addr_mod;
                    my $bopr = undef;
                    if($opr2 =~ /^\s*acc\s*$/i){
                        $addr_mod = "reg";
                        $bopr = lc($opr2);
                    }elsif($opr2 =~ /^\s*ix\s*$/i){
                        $addr_mod = "reg";
                        $bopr = lc($opr2);
                    }elsif($opr2 =~ /^\s*\[ix\s*([+-].+)\]\s*$/i){
                        $ix_flg = 1;
                        $addr_mod = "prog";
                        $bopr = [gimm($1)];
                    }elsif($opr2 =~ /^\s*\(ix\s*([+-].+)\)\s*$/i){
                        $ix_flg = 1;
                        $addr_mod = "dat";
                        $bopr = [gimm($1)];
                    }elsif($opr2 =~ /^\s*\[(.+)\]\s*$/){
                        $addr_mod = "prog";
                        $bopr = [gimm($1)];
                    }elsif($opr2 =~ /^\s*\((.+)\)\s*$/){
                        $addr_mod = "dat";
                        $bopr = [gimm($1)];
                    }else{
                        $addr_mod = "imm";
                        $bopr = [gimm($opr2)];
                    }

                    push @opr_list, {type=> "ld",
                                     op  => $opname,
                                     dst => $opr1,
                                     src => {type=>$addr_mod,
                                             ix=>$ix_flg, val=>$bopr},
                                     on  => int(@opr_list)};
                    if($addr_mod eq "reg"){
                        $caddr += 1;
                    }else{
                        $caddr += 2;
                    }
                }elsif(grep{/^$opname$/}(@shift_fmly)){
                    die if(!defined($opr1) || defined($opr2));
                    $opr1 =~ s/\s//g;
                    $opr1 = lc($opr1);
                    push @opr_list, {type=> "shift",
                                     op  => $opname,
                                     dst => $opr1,
                                     on  => int(@opr_list)};
                    $caddr += 1;
                }elsif(grep{/^$opname$/}(@branch_fmly)){
                    die if(!defined($opr1) || defined($opr2));
                    push @opr_list, {type=> "branch",
                                     op  => $opname,
                                     dst => [gimm($opr1)],
                                     on  => int(@opr_list)};
                    $caddr += 2;
                }elsif(defined($others{$opname})){
                    if($opname ne "nop"){
                        push @opr_list, {type=> "other",
                                         op  => $opname,
                                         on  => int(@opr_list)};
                    }
                    $caddr += 1;
                }elsif($opname ne "equ"){
                    die;
                }
            }
        }else{
            die;
        }
    }

    die if($caddr > 0x100);

    return (\%labels, \@opr_list);
}


my @al;
while(<>){
    push @al, $_;
}
my ($lbls, $oprs) = asm(@al);

for(keys(%$lbls)){
    $lbls->{$_}{from} = [];
    if(defined($lbls->{$_}{on})){
        $oprs->[$lbls->{$_}{on}]{jdst} = 1;
    }
}

my $opr;
for $opr(@$oprs){
    if($opr->{type} eq "branch"){
        if(@{$opr->{dst}} == 1 && $opr->{dst}[0] !~ /^\d+$/){
            push(@{$lbls->{$opr->{dst}[0]}{from}}, $opr->{on});
        }
    }
    if($opr->{type} eq "ld"){
        if($opr->{src}{type} eq "reg"){
            $opr->{src}{tval} = $opr->{src}{val};
        }else{
            my $lname;
            my $lindex = 0;
            my $idx = 0;
            for(@{$opr->{src}{val}}){
                if($_ =~ /^\d+$/){
                    $idx += $_;
                }elsif($_ =~ /^-(.+)$/){
                    $idx += $lbls->{$1}{addr};
                }else{
                    if(defined($lname)){
                        $idx += $lbls->{$_}{addr};
                    }else{
                        $lname = $_;
                        $lindex = $lbls->{$_}{addr};
                    }
                }
            }
            if($opr->{src}{type} eq "imm"){
                $opr->{src}{tval} = $idx + $lindex;
            }else{
                $opr->{src}{nval} = sprintf("%02XH", ($idx + $lindex) & 0xff);
                if($opr->{src}{type} eq "dat"){
                    $opr->{src}{nval} = "1" . $opr->{src}{nval};
                }
                if($opr->{src}{ix}){
                    if($idx){
                        $idx = "IX+$idx";
                    }else{
                        $idx = "IX";
                    }
                    $opr->{src}{nval} .= "+IX";
                }
                if(defined($lname) && $idx){
                    $opr->{src}{tval} = "${lname}[${idx}]";
                }elsif(defined($lname)){
                    $opr->{src}{tval} = $lname;
                }else{
                    if($opr->{src}{type} eq "dat"){
                        $opr->{src}{tval} = "DATA[$idx]";
                    }else{
                        $opr->{src}{tval} = "PROG[$idx]";
                    }
                }
            }
        }
    }
}

my @hoprs;
my %flgstat = (
    Z => {type=>"none"},
    N => {type=>"none"},
    V => {type=>"none"},
    C => {type=>"none"}
);

sub flgclear{
    my $reg = shift;
    my $k;
    for $k(qw/Z N V C/){
        if($flgstat{$k}{type} eq "val"){
            if($flgstat{$k}{val} eq $reg){
                $flgstat{$k}{type} = "none";
            }
        }elsif($flgstat{$k}{type} eq "cmp"){
            if($flgstat{$k}{c1} eq $reg || $flgstat{$k}{c2} eq $reg){
                $flgstat{$k}{type} = "none";
            }
        }
    }
}

my %shiftcond = (dst => "none");
for $opr(@$oprs){
    if($opr->{jdst}){
        %flgstat = (Z => {type=>"none"},
                    N => {type=>"none"},
                    V => {type=>"none"},
                    C => {type=>"none"});
        %shiftcond = (dst => "none");
    }
    
    if($opr->{type} ne "shift"){
        %shiftcond = (dst => "none");
    }

    if($opr->{type} eq "ld"){
        if($opr->{op} eq "ld"){
            flgclear($opr->{dst});
            push(@hoprs, {
                type => "std",
                text => $opr->{dst} . $st_sign . $opr->{src}{tval},
                acc  => $opr->{src}{nval}
            });
        }elsif($opr->{op} eq "st"){
            push(@hoprs, {
                type => "std",
                text => $opr->{src}{tval} . $st_sign . $opr->{dst},
                acc  => $opr->{src}{nval},
                stacc=> 1
            });
        }elsif($opr->{op} eq "cmp"){
            push(@hoprs, {
                type => "std",
                text => $opr->{dst} . " - " . $opr->{src}{tval},
                acc  => $opr->{src}{nval}
            });
            $flgstat{Z}{type} = $flgstat{N}{type}
                              = $flgstat{V}{type} = "cmp";
            $flgstat{Z}{c1} = $flgstat{N}{c1}
                            = $flgstat{V}{c1} = $opr->{dst};
            $flgstat{Z}{c2} = $flgstat{N}{c2}
                            = $flgstat{V}{c2} = $opr->{src}{tval};
            $flgstat{Z}{cmp} = $flgstat{N}{cmp}
                             = $flgstat{V}{cmp} = $hoprs[$#hoprs];
        }elsif($opr->{op} eq "sbc"){
            flgclear($opr->{dst});
            push(@hoprs, {
                type => "std",
                text => "C," . $opr->{dst} . $st_sign .
                            $opr->{dst} . "-" . $opr->{src}{tval} . "-C",
                acc  => $opr->{src}{nval}
            });
            $flgstat{Z}{type} = $flgstat{N}{type} = $flgstat{V}{type}
                              = $flgstat{C}{type} = "val";
            $flgstat{Z}{val} = $flgstat{N}{val} = $flgstat{V}{val}
                             = $flgstat{C}{val} = $opr->{dst};
        }elsif($opr->{op} eq "adc"){
            flgclear($opr->{dst});
            push(@hoprs, {
                type => "std",
                text => "C," . $opr->{dst} . $st_sign .
                            $opr->{dst} . "+" . $opr->{src}{tval} . "+C",
                acc  => $opr->{src}{nval}
            });
            $flgstat{Z}{type} = $flgstat{N}{type} = $flgstat{V}{type}
                              = $flgstat{C}{type} = "val";
            $flgstat{Z}{val} = $flgstat{N}{val} = $flgstat{V}{val}
                             = $flgstat{C}{val} = $opr->{dst};
        }elsif($opr->{op} eq "sub"){
            flgclear($opr->{dst});
            push(@hoprs, {
                type => "std",
                text => $opr->{dst} . $st_sign .
                            $opr->{dst} . "-" . $opr->{src}{tval},
                acc  => $opr->{src}{nval}
            });
            $flgstat{Z}{type} = $flgstat{N}{type}
                              = $flgstat{V}{type} = "val";
            $flgstat{Z}{val} = $flgstat{N}{val}
                             = $flgstat{V}{val} = $opr->{dst};
        }elsif($opr->{op} eq "add"){
            flgclear($opr->{dst});
            push(@hoprs, {
                type => "std",
                text => $opr->{dst} . $st_sign .
                            $opr->{dst} . "+" . $opr->{src}{tval},
                acc  => $opr->{src}{nval}
            });
            $flgstat{Z}{type} = $flgstat{N}{type}
                              = $flgstat{V}{type} = "val";
            $flgstat{Z}{val} = $flgstat{N}{val}
                             = $flgstat{V}{val} = $opr->{dst};
        }else{
            flgclear($opr->{dst});
            my %top = (eor=>"^", or=>"|", and=>"&");
            push(@hoprs, {
                type => "std",
                text => $opr->{dst} . $st_sign .  $opr->{dst}
                        . $top{$opr->{op}} . $opr->{src}{tval},
                acc  => $opr->{src}{nval}
            });
            $flgstat{V}{type} = "none";
            $flgstat{Z}{type} = $flgstat{N}{type} = "val";
            $flgstat{Z}{val} = $flgstat{N}{val} = $opr->{dst};
        }
    }elsif($opr->{type} eq "shift"){
        flgclear($opr->{dst});
        my %ss = (
            sra => {d=>">>"},
            sla => {d=>"<<"},
            srl => {d=>">>"},
            sll => {d=>"<<"},
            rra => {d=>">>", c=>1},
            rla => {d=>"<<", c=>1},
            rrl => {d=>"rot>"},
            rll => {d=>"rot<"}
        );
        my $dst = $opr->{dst};
        my $shiftnum = 1;
        if($shiftcond{dst} eq $dst && $shiftcond{op} eq $opr->{op}
                                   && $shorten_shift){
            $shiftnum = $shiftcond{num} + 1;
            $shiftcond{last}{hide} = 1;
        }
        %flgstat = (Z => {type=>"val", val=>$dst},
                    N => {type=>"val", val=>$dst},
                    V => {type=>"none"},
                    C => {type=>"none"});
        my $src = $dst;
        my $hro = $ss{$opr->{op}};
        my $left = ($hro->{d} =~ /</);
        if($hro->{c}){
            if($left){
                $src = "($src,C)";
            }else{
                $src = "(C,$src)";
            }
        }
        push(@hoprs, {
            type => "std",
            text => $dst . $st_sign .  $src . $hro->{d} . $shiftnum
        });
        %shiftcond = (
            dst => $opr->{dst},
            op  => $opr->{op},
            num => $shiftnum,
            last=> $hoprs[$#hoprs]
        );
    }elsif($opr->{type} eq "branch"){
        my $dst = undef;
        if(@{$opr->{dst}} == 1 && $opr->{dst}[0] !~ /^\d+$/){
            $dst = $opr->{dst}[0];
        }
        if($opr->{op} eq "ba"){
            push(@hoprs, {
                type => "jump",
                dst  => $dst
            });
        }else{
            my %brc = (
                bvf => {
                    flags => [qw/V/],
                    val   => undef,
                    cmp   => undef,
                    cond  => ["V", 0]
                },
                bnz => {
                    flags => [qw/Z/],
                    val   => ["=0", 1],
                    cmp   => ["=", 1],
                    cond  => ["Z", 1]
                },
                bz => {
                    flags => [qw/Z/],
                    val   => ["=0", 0],
                    cmp   => ["=", 0],
                    cond  => ["Z", 0]
                },
                bzp => {
                    flags => [qw/N/],
                    val   => ["<0", 1],
                    cmp   => undef,
                    cond  => ["N", 1]
                },
                bn => {
                    flags => [qw/N/],
                    val   => ["<0", 0],
                    cmp   => undef,
                    cond  => ["N",  0]
                },
                bp => {
                    flags => [qw/N Z/],
                    val   => [">0", 0],
                    cmp   => undef,
                    cond  => ["N|Z", 1]
                },
                bzn => {
                    flags => [qw/N Z/],
                    val   => [">0", 1],
                    cmp   => undef,
                    cond  => ["N|Z", 0]
                },
                bni => {
                    flags => [qw//],
                    val   => undef,
                    cmp   => undef,
                    cond  => ["IBUF_FLG", 1]
                },
                bno => {
                    flags => [qw//],
                    val   => undef,
                    cmp   => undef,
                    cond  => ["OBUF_FLG", 1]
                },
                bnc => {
                    flags => [qw/C/],
                    val   => undef,
                    cmp   => undef,
                    cond  => ["C", 1]
                },
                bc => {
                    flags => [qw/C/],
                    val   => undef,
                    cmp   => undef,
                    cond  => ["C", 0]
                },
                bge => {
                    flags => [qw/V N/],
                    val   => undef,
                    cmp   => [">=", 0],
                    cond  => ["V^N", 1]
                },
                blt => {
                    flags => [qw/V N/],
                    val   => undef,
                    cmp   => ["<", 0],
                    cond  => ["V^N", 0]
                },
                bgt => {
                    flags => [qw/V N Z/],
                    val   => undef,
                    cmp   => [">", 0],
                    cond  => ["(V^N)|Z", 1]
                },
                ble => {
                    flags => [qw/V N Z/],
                    val   => undef,
                    cmp   => ["<=", 0],
                    cond  => ["(V^N)|Z", 0]
                }
            );
            my %bro = %{$brc{$opr->{op}}};
            my $ftype;
            my $fvals;
            for(@{$bro{flags}}){
                if($ftype){
                    if($ftype ne $flgstat{$_}{type}){
                        $ftype = "none";
                    }
                    if($ftype eq "val" && $fvals ne $flgstat{$_}{val}){
                        $ftype = "none";
                    }
                }else{
                    $ftype = $flgstat{$_}{type};
                    $fvals = $flgstat{$_}{val};
                }
            }
            my $cond;
            my $inv;
            my $acc = undef;
            if($bro{$ftype}){
                if($ftype eq "cmp"){
                    $cond = $flgstat{$bro{flags}[0]}{c1} . $bro{$ftype}[0]
                          . $flgstat{$bro{flags}[0]}{c2};
                    $inv = $bro{$ftype}[1];
                    if($hide_compare){
                        $flgstat{$bro{flags}[0]}{cmp}{hide} = 1;
                        $acc = $flgstat{$bro{flags}[0]}{cmp}{acc};
                        $flgstat{$bro{flags}[0]}{cmp}{acc} = undef;
                    }
                }else{
                    $cond = $fvals . $bro{$ftype}[0];
                    $inv = $bro{$ftype}[1];
                }
            }else{
                $cond = $bro{cond}[0];
                $inv = $bro{cond}[1];
            }
            push(@hoprs, {
                type => "branch",
                dst  => $dst,
                cond => $cond,
                inv  => $inv,
                acc  => $acc
            });
        }
    }else{
        if($opr->{op} eq "hlt"){
            push(@hoprs, {
                type => "end",
                text => "stop"
            });
        }elsif($opr->{op} eq "out"){
            push(@hoprs, {
                type => "std",
                text => "OBUF" . $st_sign . "acc"
            });
        }elsif($opr->{op} eq "in"){
            flgclear("acc");
            push(@hoprs, {
                type => "std",
                text => "acc" . $st_sign . "IBUF"
            });
        }elsif($opr->{op} eq "rcf"){
            $flgstat{C}{type} = "none";
            push(@hoprs, {
                type => "std",
                text => "C" . $st_sign . "0"
            });
        }elsif($opr->{op} eq "scf"){
            $flgstat{C}{type} = "none";
            push(@hoprs, {
                type => "std",
                text => "C" . $st_sign . "1"
            });
        }else{
            die;
        }
    }
}

my @jll;
for(0 .. int(@hoprs)){
    push(@jll, 0);
}
my $maxheight = 0;
my @olbl;
for(keys(%$lbls)){
    my $lb = $lbls->{$_};
    if(!defined($lb->{on})){
        next;
    }
    $hoprs[$lb->{on}]{jdst} = $_;
    my $maxa = $lb->{on};
    my $mina = $lb->{on};
    for(@{$lb->{from}}){
        if($maxa < $_){
            $maxa = $_;
        }
        if($mina > $_){
            $mina = $_;
        }
    }
    $lb->{mina} = $mina;
    $lb->{maxa} = $maxa;
    push(@olbl, $lb);
}
for(sort{$a->{maxa}-$a->{mina} <=> $b->{maxa}-$b->{mina}}(@olbl)){
    my $lb = $_;
    my $mina = $lb->{mina};
    my $maxa = $lb->{maxa};
    my $height = 0;
    my $hmap = 0;
    for($mina .. $maxa){
        $hmap |= $jll[$_];
    }
    while($hmap & (1 << $height)){
        $height++;
    }
    for($mina .. $maxa){
        $jll[$_] |= 1 << $height;
    }
    $lb->{height} = $height;
    if($maxheight < $height){
        $maxheight = $height;
    }
}

my $box_xpos = $maxheight * $jumpline_offset + $jumpline_dist + 1;
my $harf_width = $box_width / 2;
my $harf_height = $box_height / 2;
my $box_xcenter = $box_xpos + $harf_width;
my $box_ypos = 1;
my @num2y_table;
my $svg = "";
for $opr(@hoprs){
    push(@num2y_table, $box_ypos);
    if($opr->{jdst}){
        if(defined($lbls->{$opr->{jdst}}{max})){
            $lbls->{$opr->{jdst}}{max} = $box_ypos
                if($lbls->{$opr->{jdst}}{max} < $box_ypos);
            $lbls->{$opr->{jdst}}{min} = $box_ypos
                if($lbls->{$opr->{jdst}}{min} > $box_ypos);
        }else{
            $lbls->{$opr->{jdst}}{max} = $box_ypos;
            $lbls->{$opr->{jdst}}{min} = $box_ypos;
        }
        my $ls = $lbls->{$opr->{jdst}}{height} * $jumpline_offset
               + $harf_width + $jumpline_dist;
        my $ss = $box_xcenter - $ls;
        my $ad = $box_xcenter - 3;
        $svg .= "<path d=\"";
        $svg .= "M $box_xcenter $box_ypos v $box_dist";
        $svg .= "\" />";
        $svg .= "<path marker-end=\"url(#Arrow)\" d=\"";
        $svg .= "M $ss $box_ypos H $ad";
        $svg .= "\" />";
        $box_ypos += $box_dist;
    }
    if($opr->{hide}){
        next;
    }
    if($opr->{acc}){
        my $tw = $access_width - $box_height/2;
        $svg .= svg_tex($opr->{acc}, $box_xpos + $box_width
                        + $access_dist + $tw/2,
                        $box_ypos + $harf_height, $tw);
        $svg .= "<path d=\"";
        $svg .= "M " . ($box_xpos + $box_width + $access_dist) . " $box_ypos "
              . "m " . ($box_height/2) . " 0 "
              . "h " . ($access_width - $box_height/2) . " "
              . "a " . (($box_height/2) . " ")x2 . "0 0 0 0 $box_height "
              . "h " . (-$access_width + $box_height/2) . " "
              . "a " . (($box_height/2) . " ")x2 . "0 0 1 0 -$box_height z";
        $svg .= "\" />";
        my $from;
        my $y = $box_ypos + $harf_height;
        my $hg = $access_dist-3;
        if(!$opr->{stacc}){
            $from = $box_xpos + $box_width + $access_dist;
            $hg = "-" . $hg;
        }else{
            $from = $box_xpos + $box_width;
        }
        $svg .= "<path marker-end=\"url(#Arrow)\" d=\"";
        $svg .= "M $from $y h $hg";
        $svg .= "\" />";
    }
    if($opr->{type} eq "std"){
        $svg .= svg_tex($opr->{text}, $box_xcenter, $box_ypos + $harf_height,
                        $box_width);
        $svg .= "<path d=\"";
        $svg .= "M $box_xpos $box_ypos "
              . "h $box_width v $box_height h -$box_width v -$box_height z";
        $svg .= "\" />";
        $box_ypos += $box_height;
        $svg .= "<path d=\"";
        $svg .= "M $box_xcenter $box_ypos v $box_dist";
        $svg .= "\" />";
        $box_ypos += $box_dist;
    }elsif($opr->{type} eq "jump"){
        my $ls = $harf_width + $jumpline_dist;
        if(defined($opr->{dst})){
            if(defined($lbls->{$opr->{dst}}{max})){
                $lbls->{$opr->{dst}}{max} = $box_ypos + $jump_preline
                    if($lbls->{$opr->{dst}}{max} < $box_ypos + $jump_preline);
                $lbls->{$opr->{dst}}{min} = $box_ypos + $jump_preline
                    if($lbls->{$opr->{dst}}{min} > $box_ypos + $jump_preline);
            }else{
                $lbls->{$opr->{dst}}{max} = $box_ypos + $jump_preline;
                $lbls->{$opr->{dst}}{min} = $box_ypos + $jump_preline;
            }
            $ls += $lbls->{$opr->{dst}}{height} * $jumpline_offset;
        }
        $svg .= "<path d=\"";
        $svg .= "M $box_xcenter $box_ypos v $jump_preline h -$ls";
        $svg .= "\" />";
        $box_ypos += $box_dist + $jump_preline;
    }elsif($opr->{type} eq "branch"){
        $svg .= svg_tex($opr->{cond}, $box_xcenter, $box_ypos + $harf_height,
                        $box_width * (1 - $text_size / $box_height));
        my $ls = $jumpline_dist;
        if(defined($opr->{dst})){
            if(defined($lbls->{$opr->{dst}}{max})){
                $lbls->{$opr->{dst}}{max} = $box_ypos + $harf_height
                    if($lbls->{$opr->{dst}}{max} < $box_ypos + $harf_height);
                $lbls->{$opr->{dst}}{min} = $box_ypos + $harf_height
                    if($lbls->{$opr->{dst}}{min} > $box_ypos + $harf_height);
            }else{
                $lbls->{$opr->{dst}}{max} = $box_ypos + $harf_height;
                $lbls->{$opr->{dst}}{min} = $box_ypos + $harf_height;
            }
            $ls += $lbls->{$opr->{dst}}{height} * $jumpline_offset;
        }
        my $igt = "NO";
        my $brt = "YES";
        if($opr->{inv}){
            $igt = "YES";
            $brt = "NO";
        }
        $svg .= "<text x=\"$box_xpos\" y=\"" . ($box_ypos+$harf_height) . "\" "
              . "font-family=\"$text_font\" font-size=\"$text_size\" "
              . "fill=\"black\" stroke=\"none\" "
              . "text-anchor=\"end\" dominant-baseline=\"text-before-edge\">"
              . "$brt</text>";
        $svg .= "<path d=\"";
        $svg .= "M $box_xpos " . ($box_ypos + $harf_height) . " h -$ls";
        $svg .= "\" />";
        $svg .= "<path d=\"";
        $svg .= "M $box_xcenter $box_ypos "
              . "l $harf_width $harf_height l -$harf_width $harf_height "
              . "l -$harf_width -$harf_height l $harf_width -$harf_height z";
        $svg .= "\" />";
        $box_ypos += $box_height;
        $svg .= "<path d=\"";
        $svg .= "M $box_xcenter $box_ypos v $text_size";
        $svg .= "\" />";
        $svg .= "<text x=\"" . ($box_xcenter+1) . "\" y=\"$box_ypos\" "
              . "font-family=\"$text_font\" font-size=\"$text_size\" "
              . "fill=\"black\" stroke=\"none\" "
              . "text-anchor=\"start\" dominant-baseline=\"text-before-edge\">"
              . "$igt</text>";
        $box_ypos += $text_size;
    }elsif($opr->{type} eq "end"){
        $svg .= svg_tex($opr->{text}, $box_xcenter, $box_ypos + $harf_height,
                        $box_width);
        $svg .= "<path d=\"";
        $svg .= "M $box_xpos $box_ypos "
              . "m " . ($box_height/2) . " 0 "
              . "h " . ($box_width - $box_height) . " "
              . "a " . (($box_height/2) . " ")x2 . "0 0 1 0 $box_height "
              . "h " . (-$box_width + $box_height) . " "
              . "a " . (($box_height/2) . " ")x2 . "0 0 1 0 -$box_height z";
        $svg .= "\" />";
        $box_ypos += $box_height;
        $box_ypos += $box_dist;
    }
}

for(keys(%$lbls)){
    my $lb = $lbls->{$_};
    if(defined($lb->{height})){
        my $xp = $box_xpos - $lb->{height} * $jumpline_offset
                           - $jumpline_dist;
        $svg .= "<path stroke-linecap=\"square\" d=\"";
        $svg .= "M $xp $lb->{min} V $lb->{max}";
        $svg .= "\" />";
    }
}

my $svg_width = $box_xpos + $box_width + $access_dist + $access_width + 1;
my $svg_height = $box_ypos + 1;

$svg = "<?xml version=\"1.0\" standalone=\"no\"?>"
     . "<svg version=\"1.1\" "
     . "baseProfile=\"full\" "
     . "width=\"" . ($svg_width * 2) . "px\" "
     . "height=\"" . ($svg_height * 2) . "px\" "
     . "viewBox=\"0 0 $svg_width $svg_height\" "
     . "xmlns=\"http://www.w3.org/2000/svg\" "
     . "xmlns:xlink=\"http://www.w3.org/1999/xlink\" "
     . "xmlns:ev=\"http://www.w3.org/2001/xml-events\">"
     . "<defs><marker id=\"Arrow\" viewBox=\"0 0 2 2\" refX=\"0\" refY=\"1\" "
     . "markerUnits=\"userSpaceOnUse\" markerWidth=\"3\" markerHeight=\"3\" "
     . "orient=\"auto\">"
     . "<path d=\"M 0 0 L 2 1 L 0 2 z\" fill=\"black\" stroke=\"none\"/>"
     . "</marker></defs>"
     . "<g fill=\"none\" stroke=\"black\" stroke-width=\"1\">"
     . $svg . "</g></svg>";

print $svg;

